/** @file Xlsx.h
    @brief Contains the interface to the Xlsx class

    This file is part of XLSX
    Output Office Open XML (OpenXML) spreadsheet files from native C++
    @author (C) Copyright 2013 Paolo Greppi libpf.com
    @date: 20130118
    @version: 0.1

    no warranties whatsoever
    distribute freely and free of charge citing this:
      http://wp.libpf.com/?p=670
 */

#ifndef XLSX_XLSX_H
#define XLSX_XLSX_H

// SYSTEM INCLUDES
#include <string>
#include <vector>
#include <map>
#include "boost/variant.hpp"

// PROJECT INCLUDES
#include <opc/opc.h>

// LOCAL INCLUDES

// FORWARD REFERENCES

typedef boost::variant<double, std::string, int> Value; 

struct Cell {
  Value value_;
  int format_;
  Cell(Value v=0, int f=0) : value_(v), format_(f) { }
}; // struct Cell

typedef std::map<int, Cell> Row;

class Sheet {
public:
  explicit Sheet(std::string name);
  /// set the cell (i, j) to the value v with format f
  /// @param i row, 0-based, from 0 to 1048576-1
  /// @param j column, 0-base, from 0 to 16384-1
  /// @param v value can be int, double or string
  /// @param f format
  /// Supported formats:
  ///   1: bold
  ///   2: bold "yyyy/mm/dd"
  ///   3: plain
  ///   4: "0.00"
  void set(int i, int j, Value v, int f);
  /// @returns sheet name
  const std::string &name(void) const;
  /// sets split panes
  /// @param i row, 0-based, from 0 to 1048576-1; set to -1 to NOT split horizontally
  /// @param j column, 0-base, from 0 to 16384-1; set to -1 to NOT split vertically
  void setSplit(int i, int j);
  /// set custom column width for currently present columns
  /// overwrites any custom column width previously set for a specific column
  /// @param w width measured as the number of characters of the maximum digit width as rendered in the normal style, or 0 to set default
  void setColumnWidth(int w);
  /// set custom column width for a specific existing column
  /// @param j column, 0-base, from 0 to 16384-1
  /// @param w width measured as the number of characters of the maximum digit width as rendered in the normal style, or 0 to set default
  void setColumnWidth(int j, int w);
private:
  std::string name_;
  std::map<int, Row> rows_;
  int minRow_, maxRow_;
  // map of columns and widths; 0 indicates no custom width is set
  // at constriction is empty
  // by default, contains the range of columns with (minColumn_ -> 0, maxColumn_ -> 0)
  std::map<int, int> columns_;
  int splitI_, splitJ_;
  
  friend class Xlsx;
}; // class Sheet

class Xlsx {
private:
  std::string fileName_, appName_, appVersion_;
  
  opcContainer *c;
  
  opcPart create_docProps_app_xml(void);
  opcPart create_docProps_core_xml(void);
  opcPart create_xl_styles_xml(void);
  opcPart create_xl_workbook_xml(void);
  opcPart create_xl_worksheets_sheet_xml(int i, const Sheet &sheet);
  void generate(void);
  
  std::vector<Sheet> sheets_;
public:
  /// creates file in memory
  /// @param fileName the name of the file that will be saved by the destructor
  Xlsx(std::string fileName);
  /// on destroying object, file will be saved
  ~Xlsx(void);

  /// will be saved in the docProps/app.xml section of the XLSX file
  /// @param appName application name 
  void setAppName(const std::string &appName);
  /// will be saved in the docProps/app.xml section of the XLSX file
  /// @param appVersion application version in the form XX.YYYY where X and Y represent numerical values
  void setAppVersion(const std::string &appVersion);
  /// @returns the file name set at construction time
  const std::string &fileName(void) const;
  /// adds a worksheet with name
  void addWorksheet(std::string name);
  /// @returns a reference to the sheet s
  /// @param s sheet index, 0-base
  Sheet &sheet(int s);
}; // class Xlsx

#endif // XLSX_XLSX_H
