/** @file Xlsx.cc
    @brief Contains the implementation of the Xlsx class

    This file is part of XLSX
    Output Office Open XML (OpenXML) spreadsheet files from native C++
    @author (C) Copyright 2013 Paolo Greppi libpf.com
    @date: 20130118
    @version: 0.1

    no warranties whatsoever
    distribute freely and free of charge citing this:
      http://wp.libpf.com/?p=670
 */

// SYSTEM INCLUDES
#include <iostream>
#include <stdexcept>
#include <algorithm> // for reverse

// PROJECT INCLUDES
#include "Xlsx.h"
#include "Timing.h"

// LOCAL INCLUDES

// FORWARD REFERENCES

#if defined(__GNUC__)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsign-compare"
#endif // __GNUC__

// /////////////////////////////////////////////////////////////////////////////
// ////////////////////////////////// fstring //////////////////////////////////
// /////////////////////////////////////////////////////////////////////////////

/// utility class with variable format
class fstring : public std::string {
public:
  fstring(const char *fmt,...) {
    int n, size = 100;
    clear();
    char *p, *np;
    if ((p = static_cast<char *>(malloc(size))) == NULL) throw std::length_error(__PRETTY_FUNCTION__);
    while (true) {
      va_list ap;
      va_start(ap, fmt);
      n = vsnprintf(p, size, fmt, ap);
      if (n > -1 && n < size) {
        assign(p);
        free(p);
        va_end(ap);
        return;
      }
      if (n > -1)
        size = n + 1;
      else
        size *= 2;
      if ((np = static_cast<char *>(realloc(p, size))) == NULL) {
        free(p);
        throw std::length_error(__PRETTY_FUNCTION__);
      } else {
        p = np;
      }
    }
  }

  operator const char*(void) { return c_str(); }
}; // class fstring

// /////////////////////////////////////////////////////////////////////////////
// /////////////////////////////////// Sheet ///////////////////////////////////
// /////////////////////////////////////////////////////////////////////////////

// http://en.wikipedia.org/wiki/Microsoft_Excel#Number_of_rows_and_columns
const int rowLimit = 1048576;
const int columnLimit = 16384;

Sheet::Sheet(std::string name) : name_(name), minRow_(1048576), maxRow_(0) {
  ;
} // Sheet::Sheet

void Sheet::set(int i, int j, Value v, int f) {
  assert(i >= 0);
  assert(i < rowLimit);
  assert(j >= 0);
  assert(j < columnLimit);
  assert(f >= 1);
  assert(f <= 4);

  if (i < minRow_)
    minRow_ = i;
  if (i > maxRow_)
    maxRow_ = i;
  std::map<int,int>::iterator p(columns_.find(j));
  if (p == columns_.end()) {
    columns_[j] = 0;
  }
  rows_[i][j] = Cell(v, f);
  // TODO: use std::map::emplace as soon as it get available (gcc 4.8 ?)
  // rows_[i].emplace(std::make_pair(j, Cell(v, f)));
} // Sheet::set

const std::string &Sheet::name(void) const {
  return name_;
} // Sheet::name

void Sheet::setSplit(int i, int j) {
  splitI_ = i;
  splitJ_ = j;
} // Sheet::setSplit

void Sheet::setColumnWidth(int w) {
  assert(w >= 0);
  for (std::map<int, int>::iterator i=columns_.begin(); i != columns_.end(); ++i)
    i->second = w;
} // Sheet::setColumnWidth

/// set custom column width for a specific column
/// @param j column, 0-base, from 0 to 16384-1
/// @param w width measured as the number of characters of the maximum digit width as rendered in the normal style
void Sheet::setColumnWidth(int j, int w) {
  assert(j >= 0);
  assert(j < columnLimit);
  assert(w >= 0);
  std::map<int,int>::iterator i(columns_.find(j));
  if (i != columns_.end()) {
    i->second = w;
  }
} // Sheet::setColumnWidth
  
// /////////////////////////////////////////////////////////////////////////////
// //////////////////////////////////// Xlsx ///////////////////////////////////
// /////////////////////////////////////////////////////////////////////////////

Xlsx::Xlsx(std::string fileName) : fileName_(fileName), appName_("XLSX"), appVersion_("00.0000") {
  ;
} // Xlsx::Xlsx

void Xlsx::setAppName(const std::string &appName) {
  appName_ = appName;
} // Xlsx::setAppName

void Xlsx::setAppVersion(const std::string &appVersion) {
  appVersion_ = appVersion;
} // Xlsx::setAppVersion

const std::string &Xlsx::fileName(void) const {
  return fileName_;
} // Xlsx::fileName

void Xlsx::addWorksheet(std::string name) {
  sheets_.push_back(Sheet(name));
} // Xlsx::addWorksheet

Xlsx::~Xlsx(void) {
  if (OPC_ERROR_NONE == opcInitLibrary()) {
    c = NULL;
    if (NULL != (c = opcContainerOpen(_X(fileName().c_str()), OPC_OPEN_WRITE_ONLY, NULL, NULL))) {
      generate();
      opcContainerClose(c, OPC_CLOSE_NOW);
    }
  } else {
    std::cerr << "Error initializing OPC library\n" << std::endl;
  }
  opcFreeLibrary();
} // Xlsx::~Xlsx

Sheet &Xlsx::sheet(int s) {
  assert(s >= 0);
  assert(s < sheets_.size());
  return sheets_[s];
} // Xlsx::sheet

void writef(opcContainerOutputStream* stream, const char *s, ...) {
  va_list ap;
  va_start(ap, s);
  char buf[1024];
  int len = vsnprintf(buf, sizeof(buf), s, ap);
  opcContainerWriteOutputStream(stream, (const opc_uint8_t *)buf, len);
  va_end(ap);
} // writef

void writes(opcContainerOutputStream* stream, const char *s) {
  int const len = strlen(s);
  opcContainerWriteOutputStream(stream, (const opc_uint8_t *)s, len);
} // writes

opcPart Xlsx::create_docProps_app_xml(void) {
  opcPart ret = opcPartFind(c, _X("docProps/app.xml"), _X("application/vnd.openxmlformats-officedocument.extended-properties+xml"), 0);
  if (OPC_PART_INVALID == ret && OPC_PART_INVALID != (ret = opcPartCreate(c, _X("docProps/app.xml"), _X("application/vnd.openxmlformats-officedocument.extended-properties+xml"), 0))) {
    //adding content
    opcContainerOutputStream *out = opcContainerCreateOutputStream(c, ret, OPC_COMPRESSIONOPTION_NORMAL);
    if (NULL != out) {
      writes(out, "<Properties xmlns=\"http://schemas.openxmlformats.org/officeDocument/2006/extended-properties\" xmlns:vt=\"http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes\">\n");
      fstring appName("  <Application>%s</Application>\n", appName_.c_str());
      writes(out, appName.c_str());
      fstring appVersion("  <AppVersion>%s</AppVersion>\n", appVersion_.c_str());
      writes(out, appVersion.c_str());
      writes(out, "  <HeadingPairs>\n");
      writes(out, "    <vt:vector size=\"2\" baseType=\"variant\">\n");
      writes(out, "      <vt:variant>\n");
      writes(out, "        <vt:lpstr>Worksheets</vt:lpstr>\n");
      writes(out, "      </vt:variant>\n");
      writes(out, "      <vt:variant>\n");
      fstring nsheets("  <vt:i4>%d</vt:i4>\n", sheets_.size());
      writes(out, nsheets.c_str());
      writes(out, "      </vt:variant>\n");
      writes(out, "    </vt:vector>\n");
      writes(out, "  </HeadingPairs>\n");
      writes(out, "  <TitlesOfParts>\n");
      fstring titles("    <vt:vector size=\"%d\" baseType=\"lpstr\">\n", sheets_.size());
      writes(out, titles.c_str());
      for (int i = 0; i < sheets_.size(); ++i) {
        fstring sheetName("      <vt:lpstr>%s</vt:lpstr>\n", sheets_[i].name().c_str());
        writes(out, sheetName.c_str());
      }
      writes(out, "    </vt:vector>\n");
      writes(out, "  </TitlesOfParts>\n");
      writes(out, "  <Company>Test S.p.A.</Company>\n");
      writes(out, "</Properties>\n");
      writes(out, "");
      opcContainerCloseOutputStream(out);
    }
    // adding relations
  }
  return ret;
} // Xlsx::create_docProps_app_xml

opcPart Xlsx::create_docProps_core_xml(void) {
  opcPart ret = opcPartFind(c, _X("docProps/core.xml"), _X("application/vnd.openxmlformats-package.core-properties+xml"), 0);
  if (OPC_PART_INVALID == ret && OPC_PART_INVALID != (ret = opcPartCreate(c, _X("docProps/core.xml"), _X("application/vnd.openxmlformats-package.core-properties+xml"), 0))) {
    //adding content
    opcContainerOutputStream *out = opcContainerCreateOutputStream(c, ret, OPC_COMPRESSIONOPTION_NORMAL);
    if (NULL != out) {
      writes(out, "<cp:coreProperties xmlns:cp=\"http://schemas.openxmlformats.org/package/2006/metadata/core-properties\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:dcterms=\"http://purl.org/dc/terms/\" xmlns:dcmitype=\"http://purl.org/dc/dcmitype/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n");
      writes(out, "  <dc:title>Financial Plan</dc:title>\n");
      writes(out, "  <dc:subject>Test title</dc:subject>\n");
      writes(out, "  <dc:creator>Test Author</dc:creator>\n");
      writes(out, "  <cp:lastModifiedBy>Test Author</cp:lastModifiedBy>\n");

      Timing timing;
      fstring created("  <dcterms:created xsi:type=\"dcterms:W3CDTF\">%s</dcterms:created>\n", timing.W3CDTF().c_str());
      writes(out, created.c_str());
      fstring modified("  <dcterms:modified xsi:type=\"dcterms:W3CDTF\">%s</dcterms:modified>\n", timing.W3CDTF().c_str());
      writes(out, modified.c_str());
      writes(out, "</cp:coreProperties>\n");
      writes(out, "");
      opcContainerCloseOutputStream(out);
    }
    // adding relations
  }
  return ret;
} // Xlsx::create_docProps_core_xml

opcPart Xlsx::create_xl_styles_xml(void) {
  opcPart ret = opcPartFind(c, _X("xl/styles.xml"), _X("application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml"), 0);
  if (OPC_PART_INVALID == ret && OPC_PART_INVALID != (ret = opcPartCreate(c, _X("xl/styles.xml"), _X("application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml"), 0))) {
    //adding content
    opcContainerOutputStream *out = opcContainerCreateOutputStream(c, ret, OPC_COMPRESSIONOPTION_NORMAL);
    if (NULL != out) {
      writes(out, "<styleSheet xmlns=\"http://schemas.openxmlformats.org/spreadsheetml/2006/main\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:x14ac=\"http://schemas.microsoft.com/office/spreadsheetml/2009/9/ac\" mc:Ignorable=\"x14ac\">\n");
      writes(out, "  <numFmts count=\"1\">\n");
      writes(out, "    <numFmt numFmtId=\"164\" formatCode=\"yyyy/mm/dd\"/>\n");
      writes(out, "  </numFmts>\n");
      writes(out, "  <fonts count=\"2\" x14ac:knownFonts=\"1\">\n");
      writes(out, "    <font>\n");
      writes(out, "      <sz val=\"11\"/>\n");
      writes(out, "      <color theme=\"1\"/>\n");
      writes(out, "      <name val=\"Calibri\"/>\n");
      writes(out, "      <family val=\"2\"/>\n");
      writes(out, "      <scheme val=\"minor\"/>\n");
      writes(out, "    </font>\n");
      writes(out, "    <font>\n");
      writes(out, "      <b/>\n");
      writes(out, "      <sz val=\"11\"/>\n");
      writes(out, "      <color theme=\"1\"/>\n");
      writes(out, "      <name val=\"Calibri\"/>\n");
      writes(out, "      <family val=\"2\"/>\n");
      writes(out, "      <scheme val=\"minor\"/>\n");
      writes(out, "    </font>\n");
      writes(out, "  </fonts>\n");
      writes(out, "  <fills count=\"2\">\n");
      writes(out, "    <fill>\n");
      writes(out, "      <patternFill patternType=\"none\"/>\n");
      writes(out, "    </fill>\n");
      writes(out, "    <fill>\n");
      writes(out, "      <patternFill patternType=\"gray125\"/>\n");
      writes(out, "    </fill>\n");
      writes(out, "  </fills>\n");
      writes(out, "  <borders count=\"1\">\n");
      writes(out, "    <border>\n");
      writes(out, "      <left/>\n");
      writes(out, "      <right/>\n");
      writes(out, "      <top/>\n");
      writes(out, "      <bottom/>\n");
      writes(out, "      <diagonal/>\n");
      writes(out, "    </border>\n");
      writes(out, "  </borders>\n");
      writes(out, "  <cellStyleXfs count=\"1\">\n");
      writes(out, "    <xf numFmtId=\"0\" fontId=\"0\" fillId=\"0\" borderId=\"0\"/>\n");
      writes(out, "  </cellStyleXfs>\n");
      writes(out, "  <cellXfs count=\"5\">\n");
      writes(out, "    <xf numFmtId=\"0\" fontId=\"0\" fillId=\"0\" borderId=\"0\" xfId=\"0\"/>\n");
      writes(out, "    <xf numFmtId=\"0\" fontId=\"1\" fillId=\"0\" borderId=\"0\" xfId=\"0\" applyFont=\"1\"/>\n");
      writes(out, "    <xf numFmtId=\"164\" fontId=\"1\" fillId=\"0\" borderId=\"0\" xfId=\"0\" applyNumberFormat=\"1\" applyFont=\"1\"/>\n");
      writes(out, "    <xf numFmtId=\"0\" fontId=\"0\" fillId=\"0\" borderId=\"0\" xfId=\"0\" applyFont=\"1\"/>\n");
      writes(out, "    <xf numFmtId=\"2\" fontId=\"0\" fillId=\"0\" borderId=\"0\" xfId=\"0\" applyNumberFormat=\"1\" applyFont=\"1\"/>\n");
      writes(out, "  </cellXfs>\n");
      writes(out, "  <cellStyles count=\"1\">\n");
      writes(out, "    <cellStyle name=\"Normal\" xfId=\"0\" builtinId=\"0\"/>\n");
      writes(out, "  </cellStyles>\n");
      writes(out, "  <dxfs count=\"0\"/>\n");
      writes(out, "  <tableStyles count=\"0\" defaultTableStyle=\"TableStyleMedium2\" defaultPivotStyle=\"PivotStyleLight16\"/>\n");
      writes(out, "  <extLst>\n");
      writes(out, "    <ext xmlns:x14=\"http://schemas.microsoft.com/office/spreadsheetml/2009/9/main\" uri=\"{EB79DEF2-80B8-43e5-95BD-54CBDDF9020C}\">\n");
      writes(out, "      <x14:slicerStyles defaultSlicerStyle=\"SlicerStyleLight1\"/>\n");
      writes(out, "    </ext>\n");
      writes(out, "    <ext xmlns:x15=\"http://schemas.microsoft.com/office/spreadsheetml/2010/11/main\" uri=\"{9260A510-F301-46a8-8635-F512D64BE5F5}\">\n");
      writes(out, "      <x15:timelineStyles defaultTimelineStyle=\"TimeSlicerStyleLight1\"/>\n");
      writes(out, "    </ext>\n");
      writes(out, "  </extLst>\n");
      writes(out, "</styleSheet>\n");
      writes(out, "");
      opcContainerCloseOutputStream(out);
    }
    // adding relations
  }
  return ret;
} // Xlsx::create_xl_styles_xml

opcPart Xlsx::create_xl_workbook_xml(void) {
  opcPart ret = opcPartFind(c, _X("xl/workbook.xml"), _X("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml"), 0);
  if (OPC_PART_INVALID == ret && OPC_PART_INVALID != (ret = opcPartCreate(c, _X("xl/workbook.xml"), _X("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml"), 0))) {
    //adding content
    opcContainerOutputStream *out = opcContainerCreateOutputStream(c, ret, OPC_COMPRESSIONOPTION_NORMAL);
    if (NULL != out) {
      writes(out, "<workbook xmlns=\"http://schemas.openxmlformats.org/spreadsheetml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:x15=\"http://schemas.microsoft.com/office/spreadsheetml/2010/11/main\" mc:Ignorable=\"x15\">\n");
      writes(out, "  <sheets>\n");
      for (int i = 0; i < sheets_.size(); ++i) {
        fstring sheetName("    <sheet name=\"%s\" sheetId=\"%d\" r:id=\"rId%d\"/>\n", sheets_[i].name().c_str(), i + 1, i + 1);
        writes(out, sheetName.c_str());
      }
      writes(out, "  </sheets>\n");
      writes(out, "</workbook>\n");
      writes(out, "");
      opcContainerCloseOutputStream(out);
    }
    // adding relations
    for (int i = 0; i < sheets_.size(); ++i) {
      fstring rId("rId%d", i + 1);
      opcRelationAdd(c, ret, _X(rId.c_str()), create_xl_worksheets_sheet_xml(i, sheets_[i]), _X("http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet"));
    }
    fstring rId("rId%d", sheets_.size() + 1);
    opcRelationAdd(c, ret, _X(rId.c_str()), create_xl_styles_xml(), _X("http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles"));
  }
  return ret;
} // Xlsx::create_xl_workbook_xml

/// Input must be greater than 0 and less than or equal to 16384
/// based on doglips's solution in:
/// http://www.dicks-blog.com/archives/2004/05/21/column-numbers-to-letters/
/// @param k The col number
std::string intToColumn(int k) {
  std::string s;
  int p;
  while (k > 0) {
    p = (k - 1) % 26;
    k = (k - p) / 26;
    s.push_back(65 + p);
  }
  std::reverse(s.begin(), s.end());
  return s;
} // intToColumn

opcPart Xlsx::create_xl_worksheets_sheet_xml(int i, const Sheet &sheet) {
  fstring path("xl/worksheets/sheet%d.xml", i + 1);
  opcPart ret = opcPartFind(c, _X(path.c_str()), _X("application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml"), 0);
  if (OPC_PART_INVALID == ret && OPC_PART_INVALID != (ret = opcPartCreate(c, _X(path.c_str()), _X("application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml"), 0))) {
    //adding content
    opcContainerOutputStream *out = opcContainerCreateOutputStream(c, ret, OPC_COMPRESSIONOPTION_NORMAL);
    if (NULL != out) {
      writes(out, "<worksheet xmlns=\"http://schemas.openxmlformats.org/spreadsheetml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:x14ac=\"http://schemas.microsoft.com/office/spreadsheetml/2009/9/ac\" mc:Ignorable=\"x14ac\">\n");
      
      writes(out, "  <sheetViews>\n");
      if (i == 0) {
        writes(out, "    <sheetView tabSelected=\"1\" workbookViewId=\"0\">\n");
      } else {
        writes(out, "    <sheetView tabSelected=\"0\" workbookViewId=\"0\">\n");
      }
      if (sheet.splitI_ < 0) {
        if (sheet.splitJ_ < 0) {
          // don't split
        } else {
          // split vertically
          fstring pane("      <pane xSplit=\"%d\" topLeftCell=\"%s1\" activePane=\"topRight\" state=\"frozen\"/>\n", sheet.splitJ_+1, intToColumn(sheet.splitJ_+2).c_str());
          writes(out, pane.c_str());
        }
      } else {
        if (sheet.splitJ_ < 0) {
          // split horizontally
          fstring pane("      <pane ySplit=\"%d\" topLeftCell=\"A%d\" activePane=\"topLeft\" state=\"frozen\"/>\n", sheet.splitI_+1, sheet.splitI_+2);
          writes(out, pane.c_str());
        } else {
          // split horizontally and vertically
          fstring pane("      <pane xSplit=\"%d\" ySplit=\"%d\" topLeftCell=\"%s%d\" activePane=\"bottomRight\" state=\"frozen\"/>\n", sheet.splitI_+1, sheet.splitJ_+1, intToColumn(sheet.splitJ_+2).c_str(), sheet.splitI_+2);
          writes(out, pane.c_str());
        }
      }
      
      writes(out, "    </sheetView>\n");
      writes(out, "  </sheetViews>\n");

      // check that any width is set
      bool widthIsSet(false);
      for (std::map<int, int>::const_iterator i=sheet.columns_.begin(); i != sheet.columns_.end(); ++i) {
        if (i->second != 0) {
          widthIsSet = true;
        } // custom width set
      } // for all columns
      if (widthIsSet) {
        writes(out, "  <cols>\n");
        for (std::map<int, int>::const_iterator i=sheet.columns_.begin(); i != sheet.columns_.end(); ++i) {
          if (i->second != 0) {
            fstring col("    <col min=\"%d\" max=\"%d\" width=\"%d\" customWidth=\"1\"/>\n", i->first+1, i->first+1, i->second);
            writes(out, col.c_str());
          } // custom width set
        } // for all columns      
        writes(out, "  </cols>\n");
      } // if widthIsSet
      
      writes(out, "  <sheetData>\n");
      for (std::map<int, Row>::const_iterator i = sheet.rows_.begin(); i != sheet.rows_.end(); ++i) {
        int r = i->first;
        fstring row("    <row r=\"%d\">\n", r + 1);
        writes(out, row.c_str());
        for (std::map<int, Cell>::const_iterator j = i->second.begin(); j != i->second.end(); ++j) {
          int column = j->first;
          const Value *v = &j->second.value_;
          int format = j->second.format_;
          if (const int *vi = boost::get<const int>(v)) {
            fstring pre("      <c r=\"%s%d\" s=\"%d\">\n", intToColumn(column + 1).c_str(), r + 1, format);
            writes(out, pre.c_str());
            fstring cell("        <v>%d</v>\n", *vi);
            writes(out, cell.c_str());
            writes(out, "      </c>\n");
          } else if (const double *vd = boost::get<const double>(v)) {
            fstring pre("      <c r=\"%s%d\" s=\"%d\">\n", intToColumn(column + 1).c_str(), r + 1, format);
            writes(out, pre.c_str());
            fstring cell("        <v>%g</v>\n", *vd);
            writes(out, cell.c_str());
            writes(out, "      </c>\n");
          } else if (const std::string * vs = boost::get<const std::string>(v)) {
            fstring pre("      <c r=\"%s%d\" s=\"%d\" t=\"inlineStr\">\n", intToColumn(column + 1).c_str(), r + 1, format);
            writes(out, pre.c_str());
            fstring cell("        <is><t>%s</t></is>\n", vs->c_str());
            writes(out, cell.c_str());
            writes(out, "      </c>\n");
          }
        }
        writes(out, "    </row>\n");
      }
      writes(out, "  </sheetData>\n");

      writes(out, "</worksheet>\n");
      writes(out, "");
      opcContainerCloseOutputStream(out);
    }
  }
  return ret;
} // Xlsx::create_xl_worksheets_sheet_xml

void Xlsx::generate(void) {
  // adding registered extensions
  opcExtensionRegister(c, _X("rels"), _X("application/vnd.openxmlformats-package.relationships+xml"));
  opcExtensionRegister(c, _X("xml"), _X("application/xml"));
  // adding root relations
  opcRelationAdd(c, OPC_PART_INVALID, _X("rId1"), create_xl_workbook_xml(), _X("http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument"));
  opcRelationAdd(c, OPC_PART_INVALID, _X("rId2"), create_docProps_core_xml(), _X("http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties"));
  opcRelationAdd(c, OPC_PART_INVALID, _X("rId3"), create_docProps_app_xml(), _X("http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties"));
} // Xlsx::generate

#if defined(__GNUC__)
#pragma GCC diagnostic pop
#endif // __GNUC__
